# 🧬 QuantumLangChain

[![PyPI version](https://badge.fury.io/py/quantumlangchain.svg)](https://badge.fury.io/py/quantumlangchain)
[![Python 3.9+](https://img.shields.io/badge/python-3.9+-blue.svg)](https://www.python.org/downloads/)
[![License: Commercial](https://img.shields.io/badge/License-Commercial-red.svg)](#-licensing)
[![Documentation](https://img.shields.io/badge/docs-mkdocs-blue.svg)](https://krish567366.github.io/Quantum-Langchain)
[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)

**LICENSED SOFTWARE: A composable framework for quantum-inspired reasoning, entangled memory systems, and multi-agent cooperation — engineered for next-gen artificial intelligence.**

**📧 Contact: bajpaikrishna715@gmail.com for licensing**  
**⏰ 24-hour grace period available for evaluation**

---

## Licensing

**⚠️ IMPORTANT: QuantumLangChain is commercial software requiring a valid license for all features beyond the 24-hour evaluation period.**

### Quick Start with Licensing

1. **Install**: `pip install quantumlangchain`
2. **Import**: Automatically starts 24-hour evaluation
3. **Get Machine ID**: `python -c "import quantumlangchain; print(quantumlangchain.get_machine_id())"`
4. **Contact**: Email bajpaikrishna715@gmail.com with your machine ID
5. **Activate**: Receive and activate your license file

### License Tiers

| Tier | Price | Features | Operations/Day |
|------|-------|----------|----------------|
| 🆓 **Evaluation** | Free (24 hours) | Basic features only | 1,000 |
| 💼 **Basic** | $29/month | Core + Chains + Memory | 10,000 |
| 🚀 **Professional** | $99/month | Multi-Agent + Advanced | 100,000 |
| 🏢 **Enterprise** | $299/month | All features + Support | Unlimited |
| 🎓 **Research** | $49/month | Academic features | 50,000 |

---

## 🧬 About QuantumLangChain

QuantumLangChain bridges the gap between classical AI and quantum computing, providing a unified framework for building hybrid quantum-classical AI systems with advanced memory management, multi-agent cooperation, and quantum-inspired reasoning capabilities.

## 🚀 Features

### 🔧 Core Modules

- **QLChain**: Quantum-ready chains with decoherence-aware control flows and circuit injection
- **QuantumMemory**: Reversible, entangled memory layers with hybrid vector store support
- **QuantumToolExecutor**: Tool execution router with quantum-classical API bridge
- **EntangledAgents**: Multi-agent systems with shared memory entanglement and interference-based reasoning
- **QPromptChain**: Prompt chaining with quantum-style uncertainty branching
- **QuantumRetriever**: Quantum-enhanced semantic retrieval using Grover-based subquery refinement
- **QuantumContextManager**: Temporal snapshots and dynamic context expansion

### 🧬 Advanced Capabilities

- **Decoherence-Aware Reasoning**: Simulate quantum noise impact on logic and decision trees
- **Timeline Rewriting**: Memory snapshotting, branching, and rollback of reasoning paths
- **Entangled Collaboration**: Agents with shared belief states and quantum-style communication
- **Self-Adaptive Reasoning Graphs**: Dynamic agent chain restructuring during execution

## 📦 Installation

### Basic Installation

```bash
pip install quantumlangchain
```

### Development Installation

```bash
pip install quantumlangchain[dev]
```

### Full Installation (with all optional dependencies)

```bash
pip install quantumlangchain[all]
```

### From Source

```bash
git clone https://github.com/krish567366/Quantum-Langchain.git
cd Quantum-Langchain
pip install -e .
```

## 🧠 Quick Start

### Basic Quantum Chain

```python
from quantumlangchain import QLChain, QuantumMemory
from quantumlangchain.backends import QiskitBackend

# Initialize quantum backend
backend = QiskitBackend()

# Create quantum memory
memory = QuantumMemory(
    classical_dim=512,
    quantum_dim=8,
    backend=backend
)

# Build a quantum chain
chain = QLChain(
    memory=memory,
    decoherence_threshold=0.1,
    circuit_depth=10
)

# Execute with quantum-classical hybrid reasoning
result = await chain.arun("Analyze the quantum implications of this dataset")
```

### Multi-Agent Entanglement

```python
from quantumlangchain import EntangledAgents, SharedQuantumMemory

# Create shared quantum memory
shared_memory = SharedQuantumMemory(agents=3, entanglement_depth=4)

# Initialize entangled agents
agents = EntangledAgents(
    agent_count=3,
    shared_memory=shared_memory,
    interference_weight=0.3
)

# Collaborative quantum reasoning
results = await agents.collaborative_solve(
    "Complex multi-dimensional optimization problem"
)
```

### Quantum-Enhanced Retrieval

```python
from quantumlangchain import QuantumRetriever
from quantumlangchain.vectorstores import HybridChromaDB

# Setup hybrid vector store
vectorstore = HybridChromaDB(
    classical_embeddings=True,
    quantum_embeddings=True,
    entanglement_degree=2
)

# Quantum retriever with Grover enhancement
retriever = QuantumRetriever(
    vectorstore=vectorstore,
    grover_iterations=3,
    quantum_speedup=True
)

# Enhanced semantic search
docs = await retriever.aretrieve("quantum machine learning applications")
```

## 🛠️ Supported Quantum Backends

- **Qiskit**: IBM Quantum platform integration
- **PennyLane**: Differentiable quantum programming
- **Amazon Braket**: AWS quantum computing service
- **Cirq**: Google's quantum computing framework
- **Qulacs**: High-performance quantum simulator

## 📚 Documentation

Comprehensive documentation is available at [krish567366.github.io/Quantum-Langchain](https://krish567366.github.io/Quantum-Langchain)

### Key Sections

- [Getting Started Guide](https://krish567366.github.io/Quantum-Langchain/getting-started/)
- [API Reference](https://krish567366.github.io/Quantum-Langchain/api/)
- [Quantum Concepts](https://krish567366.github.io/Quantum-Langchain/concepts/)
- [Integration Tutorials](https://krish567366.github.io/Quantum-Langchain/tutorials/)
- [Examples Gallery](https://krish567366.github.io/Quantum-Langchain/examples/)

## 🧪 Examples

Check out our comprehensive examples in the `/examples` directory:

- **Basic Quantum Reasoning**: `examples/basic_quantum_chain.ipynb`
- **Memory Entanglement**: `examples/quantum_memory_demo.ipynb`
- **Multi-Agent Systems**: `examples/entangled_agents.ipynb`
- **Quantum Retrieval**: `examples/quantum_rag_system.ipynb`
- **Timeline Manipulation**: `examples/temporal_reasoning.ipynb`

## 🧬 Architecture

QuantumLangChain follows a modular, extensible architecture:

```bash
quantumlangchain/
├── core/           # Core quantum-classical interfaces
├── chains/         # QLChain implementations
├── memory/         # Quantum memory systems
├── agents/         # Entangled agent frameworks
├── tools/          # Quantum tool executors
├── retrievers/     # Quantum-enhanced retrieval
├── backends/       # Quantum backend abstractions
├── vectorstores/   # Hybrid vector databases
└── utils/          # Utility functions and helpers
```

## 🤝 Contributing

We welcome contributions! Please see our [Contributing Guide](CONTRIBUTING.md) for details.

### Development Setup

```bash
git clone https://github.com/krish567366/Quantum-Langchain.git
cd Quantum-Langchain
pip install -e .[dev]
pre-commit install
```

### Running Tests

```bash
pytest tests/
```

### Code Formatting

```bash
black quantumlangchain/
ruff check quantumlangchain/
```

## 📊 Performance Benchmarks

| Operation | Classical Time | Quantum-Enhanced Time | Speedup |
|-----------|---------------|----------------------|---------|
| Semantic Search | 150ms | 45ms | 3.3x |
| Multi-Agent Reasoning | 800ms | 320ms | 2.5x |
| Memory Retrieval | 100ms | 35ms | 2.9x |
| Chain Execution | 500ms | 200ms | 2.5x |

### *Benchmarks run on quantum simulators with 16 qubits*

## 🔮 Roadmap

- [ ] **Q1 2025**: Hardware quantum backend integration
- [ ] **Q2 2025**: Advanced error correction protocols
- [ ] **Q3 2025**: Quantum neural network support
- [ ] **Q4 2025**: Distributed quantum computing

## 📄 License

This project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details.

## 🙏 Acknowledgments

- Inspired by LangChain's composable AI architecture
- Built on the shoulders of giants in quantum computing
- Special thanks to the quantum computing research community

## 📞 Contact

### **Krishna Bajpai**

- Email: [bajpaikrishna715@gmail.com](mailto:bajpaikrishna715@gmail.com)
- GitHub: [@krish567366](https://github.com/krish567366)
- Project: [Quantum-Langchain](https://github.com/krish567366/Quantum-Langchain)

---

**"Bridging the quantum-classical divide in artificial intelligence"** 🌉⚛️🤖
